"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var rxjs_1 = require("rxjs");
var operators_1 = require("rxjs/operators");
var common_1 = require("../common");
var TemplateRequest_1 = require("./TemplateRequest");
var Template = (function () {
    function Template(args) {
        this._disposed$ = new rxjs_1.Subject();
        this._events$ = new rxjs_1.Subject();
        this.isDisposed = false;
        this.disposed$ = this._disposed$.pipe(operators_1.share());
        this.events$ = this._events$.pipe(operators_1.takeUntil(this.disposed$), operators_1.share());
        this.config = {
            processors: [],
            sources: [],
            filters: [],
            cache: {
                files: undefined,
            },
        };
        var sources = args.sources, filters = args.filters, processors = args.processors;
        this.config.sources = sources || this.config.sources;
        this.config.filters = filters || this.config.filters;
        this.config.processors = processors || this.config.processors;
    }
    Template.create = function (source) {
        var tmpl = new Template({});
        return source ? tmpl.add(source) : tmpl;
    };
    Template.prototype.dispose = function () {
        this.isDisposed = true;
        this._disposed$.next();
    };
    Template.prototype.clone = function (args) {
        if (args === void 0) { args = {}; }
        return new Template({
            sources: args.sources || this.config.sources,
            processors: args.processors || this.config.processors,
            filters: args.filters || this.config.filters,
        });
    };
    Object.defineProperty(Template.prototype, "sources", {
        get: function () {
            return this.config.sources;
        },
        enumerable: true,
        configurable: true
    });
    Template.prototype.add = function (source) {
        var sources = this.sources.slice();
        var list = Array.isArray(source) ? source : [source];
        list.forEach(function (item) {
            if (item instanceof Template) {
                return (sources = sources.concat(item.sources));
            }
            if (typeof item === 'string') {
                return (sources = sources.concat([{ dir: item }]));
            }
            return (sources = sources.concat([item]));
        });
        sources = common_1.R.uniq(sources);
        return this.clone({ sources: sources });
    };
    Template.prototype.filter = function (fn) {
        var filters = this.config.filters.concat([fn]);
        return this.clone({ filters: filters });
    };
    Template.prototype.use = function (arg1, arg2) {
        var fn = typeof arg2 === 'function' ? arg2 : arg1;
        if (!fn) {
            throw new Error("A template processor function must be specified");
        }
        var pathFilter = [];
        if (arg1 instanceof RegExp || Array.isArray(arg1)) {
            var list = Array.isArray(arg1) ? arg1 : [arg1];
            pathFilter = list.filter(function (filter) { return filter instanceof RegExp; });
        }
        fn.pathFilters = pathFilter;
        var processors = this.config.processors.concat([fn]);
        return this.clone({ processors: processors });
    };
    Template.prototype.files = function (options) {
        if (options === void 0) { options = {}; }
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var cache, wait, files, _a, _b, exists, filters;
            return tslib_1.__generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        cache = common_1.value.defaultValue(options.cache, true);
                        if (cache && this.config.cache.files) {
                            return [2, this.config.cache.files];
                        }
                        wait = this.sources.map(function (source) { return getFiles(source); });
                        _b = (_a = common_1.value).flatten;
                        return [4, Promise.all(wait)];
                    case 1:
                        files = _b.apply(_a, [_c.sent()]);
                        exists = function (file, list) {
                            return list.findIndex(function (f) { return f.source === file.source; }) > -1;
                        };
                        files = files
                            .reverse()
                            .reduce(function (acc, next) { return (exists(next, acc) ? acc : acc.concat([next])); }, [])
                            .reverse();
                        filters = this.config.filters;
                        files =
                            filters.length === 0
                                ? files
                                : files.filter(function (file) { return filters.every(function (filter) { return filter(file); }); });
                        this.config.cache.files = files;
                        return [2, files];
                }
            });
        });
    };
    Template.prototype.execute = function (args) {
        if (args === void 0) { args = {}; }
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var cache, _a, variables, processors, files, events$, payload, wait;
            return tslib_1.__generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        cache = args.cache, _a = args.variables, variables = _a === void 0 ? {} : _a;
                        processors = this.config.processors;
                        if (processors.length === 0) {
                            return [2];
                        }
                        return [4, this.files({ cache: cache })];
                    case 1:
                        files = _b.sent();
                        events$ = this._events$;
                        payload = { files: files };
                        events$.next({ type: 'EXECUTE/start', payload: payload });
                        wait = files.map(function (file) {
                            return runProcessors({ processors: processors, file: file, variables: variables, events$: events$ });
                        });
                        return [4, Promise.all(wait)];
                    case 2:
                        _b.sent();
                        events$.next({ type: 'EXECUTE/complete', payload: payload });
                        return [2];
                }
            });
        });
    };
    return Template;
}());
exports.Template = Template;
function getFiles(source) {
    return tslib_1.__awaiter(this, void 0, void 0, function () {
        var dir, _a, pattern, targetPath, base, formatPath, toTarget, toFile, stats, error, file, path_1, path, paths, wait;
        var _this = this;
        return tslib_1.__generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    dir = source.dir, _a = source.pattern, pattern = _a === void 0 ? '**' : _a, targetPath = source.targetDir;
                    base = common_1.fsPath.resolve(dir);
                    return [4, common_1.fs.pathExists(base)];
                case 1:
                    if (!(_b.sent())) {
                        return [2, []];
                    }
                    formatPath = function (path) { return path.substr(base.length); };
                    toTarget = function (path) {
                        path = formatPath(path);
                        path = targetPath ? common_1.fsPath.join(targetPath, path) : path;
                        return path;
                    };
                    toFile = function (path) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                        var file, _a;
                        return tslib_1.__generator(this, function (_b) {
                            switch (_b.label) {
                                case 0:
                                    _a = {
                                        base: base,
                                        source: formatPath(path),
                                        target: toTarget(path)
                                    };
                                    return [4, common_1.isBinaryFile(path)];
                                case 1:
                                    file = (_a.isBinary = _b.sent(),
                                        _a);
                                    return [2, file];
                            }
                        });
                    }); };
                    return [4, common_1.fs.lstat(base)];
                case 2:
                    stats = _b.sent();
                    if (!stats.isFile()) return [3, 4];
                    if (source.pattern) {
                        error = '';
                        error += "A full file-path was specified as the 'dir' and a glob 'pattern' was given. ";
                        error += "Can only be one of the other. ";
                        error += "\npattern: '" + source.pattern + "', \ndir: '" + source.dir + "'";
                        throw new Error(error);
                    }
                    file = common_1.fsPath.basename(base);
                    base = common_1.fsPath.dirname(base);
                    path_1 = common_1.fsPath.join(base, file);
                    return [4, toFile(path_1)];
                case 3: return [2, [_b.sent()]];
                case 4:
                    path = common_1.fsPath.join(base, pattern);
                    return [4, common_1.glob.find(common_1.fsPath.resolve(path), {
                            type: 'FILES',
                            dot: true,
                        })];
                case 5:
                    paths = _b.sent();
                    wait = paths.map(function (path) { return toFile(path); });
                    return [2, Promise.all(wait)];
            }
        });
    });
}
function runProcessors(args) {
    var _this = this;
    return new Promise(function (resolve, reject) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
        var processors, file, variables, isResolved, runComplete_1, buffer_1, text_1, res_1, runProcessor_1, index_1, runNext_1, error_1;
        var _this = this;
        return tslib_1.__generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    processors = args.processors, file = args.file, variables = args.variables;
                    isResolved = false;
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    runComplete_1 = function () {
                        if (!isResolved) {
                            isResolved = true;
                            resolve();
                        }
                    };
                    return [4, common_1.fs.readFile(common_1.fsPath.join(file.base, file.source))];
                case 2:
                    buffer_1 = _a.sent();
                    text_1 = file.isBinary ? undefined : buffer_1.toString();
                    res_1 = {
                        next: function () { return runNext_1(); },
                        complete: function () { return runComplete_1(); },
                        done: function (next) { return (next === 'COMPLETE' ? runComplete_1() : runNext_1()); },
                        get text() {
                            return text_1;
                        },
                        set text(value) {
                            text_1 = value;
                        },
                        replaceText: function (searchValue, replaceValue) {
                            text_1 =
                                text_1 === undefined
                                    ? undefined
                                    : text_1.replace(searchValue, replaceValue);
                            return res_1;
                        },
                        alert: function (payload) {
                            args.events$.next({ type: 'ALERT', payload: payload });
                            return this;
                        },
                    };
                    runProcessor_1 = function (index) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                        var fn, source, target, path, content, filters, isMatch, req;
                        return tslib_1.__generator(this, function (_a) {
                            fn = processors[index];
                            if (isResolved || !fn) {
                                return [2];
                            }
                            source = file.source, target = file.target;
                            path = { source: source, target: target };
                            content = text_1 || buffer_1;
                            filters = fn.pathFilters || [];
                            if (filters.length > 0) {
                                isMatch = filters.some(function (filter) { return filter.test(path.target); });
                                if (!isMatch) {
                                    return [2, runNext_1()];
                                }
                            }
                            req = new TemplateRequest_1.TemplateRequest({ path: path, content: content, variables: variables });
                            fn(req, res_1);
                            return [2];
                        });
                    }); };
                    index_1 = 0;
                    runNext_1 = function () {
                        index_1++;
                        if (index_1 < processors.length) {
                            runProcessor_1(index_1);
                        }
                        else {
                            runComplete_1();
                        }
                    };
                    runProcessor_1(0);
                    return [3, 4];
                case 3:
                    error_1 = _a.sent();
                    reject(error_1);
                    return [3, 4];
                case 4: return [2];
            }
        });
    }); });
}
