import { Subject } from 'rxjs';
import { share, takeUntil } from 'rxjs/operators';
import { fs, fsPath, glob, isBinaryFile, R, value } from '../common';
import { TemplateRequest } from './TemplateRequest';
export class Template {
    constructor(args) {
        this._disposed$ = new Subject();
        this._events$ = new Subject();
        this.isDisposed = false;
        this.disposed$ = this._disposed$.pipe(share());
        this.events$ = this._events$.pipe(takeUntil(this.disposed$), share());
        this.config = {
            processors: [],
            sources: [],
            filters: [],
            cache: {
                files: undefined,
            },
        };
        const { sources, filters, processors } = args;
        this.config.sources = sources || this.config.sources;
        this.config.filters = filters || this.config.filters;
        this.config.processors = processors || this.config.processors;
    }
    static create(source) {
        const tmpl = new Template({});
        return source ? tmpl.add(source) : tmpl;
    }
    dispose() {
        this.isDisposed = true;
        this._disposed$.next();
    }
    clone(args = {}) {
        return new Template({
            sources: args.sources || this.config.sources,
            processors: args.processors || this.config.processors,
            filters: args.filters || this.config.filters,
        });
    }
    get sources() {
        return this.config.sources;
    }
    add(source) {
        let sources = [...this.sources];
        const list = Array.isArray(source) ? source : [source];
        list.forEach(item => {
            if (item instanceof Template) {
                return (sources = [...sources, ...item.sources]);
            }
            if (typeof item === 'string') {
                return (sources = [...sources, { dir: item }]);
            }
            return (sources = [...sources, item]);
        });
        sources = R.uniq(sources);
        return this.clone({ sources });
    }
    filter(fn) {
        const filters = [...this.config.filters, fn];
        return this.clone({ filters });
    }
    use(arg1, arg2) {
        const fn = typeof arg2 === 'function' ? arg2 : arg1;
        if (!fn) {
            throw new Error(`A template processor function must be specified`);
        }
        let pathFilter = [];
        if (arg1 instanceof RegExp || Array.isArray(arg1)) {
            const list = Array.isArray(arg1) ? arg1 : [arg1];
            pathFilter = list.filter(filter => filter instanceof RegExp);
        }
        fn.pathFilters = pathFilter;
        const processors = [...this.config.processors, fn];
        return this.clone({ processors });
    }
    async files(options = {}) {
        const cache = value.defaultValue(options.cache, true);
        if (cache && this.config.cache.files) {
            return this.config.cache.files;
        }
        const wait = this.sources.map(source => getFiles(source));
        let files = value.flatten(await Promise.all(wait));
        const exists = (file, list) => list.findIndex(f => f.source === file.source) > -1;
        files = files
            .reverse()
            .reduce((acc, next) => (exists(next, acc) ? acc : [...acc, next]), [])
            .reverse();
        const filters = this.config.filters;
        files =
            filters.length === 0
                ? files
                : files.filter(file => filters.every(filter => filter(file)));
        this.config.cache.files = files;
        return files;
    }
    async execute(args = {}) {
        const { cache, variables = {} } = args;
        const processors = this.config.processors;
        if (processors.length === 0) {
            return;
        }
        const files = await this.files({ cache });
        const events$ = this._events$;
        const payload = { files };
        events$.next({ type: 'EXECUTE/start', payload });
        const wait = files.map(file => runProcessors({ processors, file, variables, events$ }));
        await Promise.all(wait);
        events$.next({ type: 'EXECUTE/complete', payload });
    }
}
async function getFiles(source) {
    const { dir, pattern = '**', targetDir: targetPath } = source;
    let base = fsPath.resolve(dir);
    if (!(await fs.pathExists(base))) {
        return [];
    }
    const formatPath = (path) => path.substr(base.length);
    const toTarget = (path) => {
        path = formatPath(path);
        path = targetPath ? fsPath.join(targetPath, path) : path;
        return path;
    };
    const toFile = async (path) => {
        const file = {
            base,
            source: formatPath(path),
            target: toTarget(path),
            isBinary: await isBinaryFile(path),
        };
        return file;
    };
    const stats = await fs.lstat(base);
    if (stats.isFile()) {
        if (source.pattern) {
            let error = '';
            error += `A full file-path was specified as the 'dir' and a glob 'pattern' was given. `;
            error += `Can only be one of the other. `;
            error += `\npattern: '${source.pattern}', \ndir: '${source.dir}'`;
            throw new Error(error);
        }
        const file = fsPath.basename(base);
        base = fsPath.dirname(base);
        const path = fsPath.join(base, file);
        return [await toFile(path)];
    }
    const path = fsPath.join(base, pattern);
    const paths = await glob.find(fsPath.resolve(path), {
        type: 'FILES',
        dot: true,
    });
    const wait = paths.map(path => toFile(path));
    return Promise.all(wait);
}
function runProcessors(args) {
    return new Promise(async (resolve, reject) => {
        const { processors, file, variables } = args;
        let isResolved = false;
        try {
            const runComplete = () => {
                if (!isResolved) {
                    isResolved = true;
                    resolve();
                }
            };
            const buffer = await fs.readFile(fsPath.join(file.base, file.source));
            let text = file.isBinary ? undefined : buffer.toString();
            const res = {
                next: () => runNext(),
                complete: () => runComplete(),
                done: next => (next === 'COMPLETE' ? runComplete() : runNext()),
                get text() {
                    return text;
                },
                set text(value) {
                    text = value;
                },
                replaceText(searchValue, replaceValue) {
                    text =
                        text === undefined
                            ? undefined
                            : text.replace(searchValue, replaceValue);
                    return res;
                },
                alert(payload) {
                    args.events$.next({ type: 'ALERT', payload });
                    return this;
                },
            };
            const runProcessor = async (index) => {
                const fn = processors[index];
                if (isResolved || !fn) {
                    return;
                }
                const { source, target } = file;
                const path = { source, target };
                const content = text || buffer;
                const filters = fn.pathFilters || [];
                if (filters.length > 0) {
                    const isMatch = filters.some(filter => filter.test(path.target));
                    if (!isMatch) {
                        return runNext();
                    }
                }
                const req = new TemplateRequest({ path, content, variables });
                fn(req, res);
            };
            let index = 0;
            const runNext = () => {
                index++;
                if (index < processors.length) {
                    runProcessor(index);
                }
                else {
                    runComplete();
                }
            };
            runProcessor(0);
        }
        catch (error) {
            reject(error);
        }
    });
}
